import os
import shutil
import zipfile
from functools import reduce
from pathlib import Path
from textwrap import dedent
from typing import Optional, Tuple

from dl_plus.cli.args import dlp_config as dlp_config_arg
from dl_plus.config import Config
from dl_plus.exceptions import DLPlusException
from dl_plus.pypi import PyPIClient, Wheel, load_metadata, save_metadata


class _CommandBase(object):

    def __init_subclass__(cls):
        dct = cls.__dict__
        if 'name' not in dct:
            setattr(cls, 'name', dct['__module__'].rpartition('.')[-1])
        long_description = dct.get('long_description')
        if long_description:
            long_description = dedent(long_description).rstrip() + '\n'
            setattr(cls, 'long_description', long_description)


class Command(_CommandBase):

    name = None
    short_description = None
    long_description = None
    arguments = ()

    config = None

    def __init__(self, args) -> None:
        self.args = args
        if dlp_config_arg in self.arguments:
            config = Config()
            if not args.no_dlp_config:
                config.load(args.dlp_config)
            self.config = config
        self.init()

    def init(self) -> None:
        pass

    def run(self) -> None:
        raise NotImplementedError


class CommandGroup(_CommandBase):

    name = None
    short_description = None
    long_description = None
    commands = ()

    def __init_subclass__(cls):
        super().__init_subclass__()
        cls._commands = {cmd.name: cmd for cmd in cls.commands}

    @classmethod
    def get_command(cls, command_path):
        return reduce(lambda c, p: c._commands[p], command_path, cls)


class CommandError(DLPlusException):

    pass


class BaseInstallCommand(Command):

    def get_output_dir(self, wheel: Wheel) -> Path:
        raise NotImplementedError

    def get_name_version_tuple(self) -> Tuple[str, Optional[str]]:
        raise NotImplementedError

    def get_force_flag(self) -> bool:
        return False

    def run(self):
        client = PyPIClient()
        name, version = self.get_name_version_tuple()
        wheel = client.fetch_wheel_info(name, version)
        print(f'Found remote version: {wheel.name} {wheel.version}')
        output_dir = self.get_output_dir(wheel)
        if output_dir.exists():
            installed_metadata = load_metadata(output_dir)
            if installed_metadata:
                print(
                    f'Found installed version: {installed_metadata.name} '
                    f'{installed_metadata.version}'
                )
                if installed_metadata.version == wheel.version:
                    print('The same version is already installed')
                    if not self.args.force:
                        print('Nothing to do')
                        return
                    print('Forcing installation')
            shutil.rmtree(output_dir)
        print('Installing')
        os.makedirs(output_dir)
        with client.download_file(wheel.url, wheel.sha256) as fobj:
            with zipfile.ZipFile(fobj) as zfobj:
                zfobj.extractall(output_dir)
        save_metadata(output_dir, wheel.metadata)
        print('Done')



from .backend import BackendCommandGroup
from .base import CommandGroup
from .extractor import ExtractorCommandGroup


class RootCommandGroup(CommandGroup):

    short_description = 'commands'

    commands = (
        BackendCommandGroup,
        ExtractorCommandGroup,
    )



import hashlib
import json
from collections import namedtuple
from io import BytesIO
from pathlib import Path
from typing import Dict, Optional
from urllib.error import HTTPError
from urllib.request import urlopen

from dl_plus.exceptions import DLPlusException


Wheel = namedtuple('Wheel', 'name,version,metadata,filename,url,sha256')


class PyPIClientError(DLPlusException):

    pass


class RequestError(PyPIClientError):

    pass


class ParseError(PyPIClientError):

    pass


class DownloadError(RequestError):

    def __init__(
        self, project_name: str, version: Optional[str], error: str,
    ) -> None:
        if version:
            message = f'{project_name} {version}: {error}'
        else:
            message = f'{project_name}: {error}'
        super().__init__(message)


class Metadata(dict):

    @property
    def name(self) -> str:
        return self['info']['name']

    @property
    def version(self) -> str:
        return self['info']['version']

    @property
    def urls(self) -> Dict[str, Dict]:
        return self['urls']


def save_metadata(backend_dir: Path, metadata: Metadata) -> None:
    with open(backend_dir / 'metadata.json', 'w') as fobj:
        json.dump(metadata, fobj)


def load_metadata(backend_dir: Path) -> Metadata:
    try:
        with open(backend_dir / 'metadata.json') as fobj:
            return Metadata(json.load(fobj))
    except OSError:
        return None


class PyPIClient:

    JSON_BASE_URL = 'https://pypi.org/pypi'

    def build_json_url(
        self, project_name: str, version: Optional[str] = None,
    ) -> str:
        parts = [self.JSON_BASE_URL, project_name]
        if version:
            parts.append(version)
        parts.append('json')
        return '/'.join(parts)

    def fetch_metadata(
        self, project_name: str, version: Optional[str] = None,
    ) -> Metadata:
        url = self.build_json_url(project_name, version)
        try:
            with urlopen(url) as response:
                return Metadata(json.load(response))
        except (OSError, ValueError) as exc:
            raise RequestError from exc

    def _is_wheel_release(self, release: Dict) -> bool:
        return (
            release['packagetype'] == 'bdist_wheel' and not release['yanked'])

    def fetch_wheel_info(
        self, project_name: str, version: Optional[str] = None,
    ) -> Wheel:
        try:
            metadata = self.fetch_metadata(project_name, version)
        except RequestError as exc:
            if isinstance(exc.error, HTTPError) and exc.error.code == 404:
                error = 'not found'
            else:
                error = 'unexpected error'
            raise DownloadError(project_name, version, error) from exc
        try:
            release = next(filter(self._is_wheel_release, metadata.urls))
        except StopIteration:
            raise DownloadError(project_name, version, 'no wheel distribution')
        return Wheel(
            name=metadata.name,
            version=metadata.version,
            metadata=metadata,
            filename=release['filename'],
            url=release['url'],
            sha256=release['digests']['sha256'],
        )

    def download_file(self, url: str, sha256: Optional[str] = None) -> BytesIO:
        try:
            with urlopen(url) as response:
                buffer = BytesIO(response.read())
        except OSError as exc:
            raise DownloadError(f'{url}: {exc}') from exc
        if sha256:
            digest = hashlib.sha256(buffer.getvalue())
            hexdigest = digest.hexdigest()
            if hexdigest != sha256:
                raise DownloadError(
                    f'{url}: sha256 mismatch: expected {sha256}, '
                    f'got {hexdigest}'
                )
        return buffer



from pathlib import Path
from typing import Optional, Tuple

from dl_plus import backend
from dl_plus.cli.args import Arg
from dl_plus.cli.commands.base import BaseInstallCommand
from dl_plus.pypi import Wheel


class BackendInstallCommand(BaseInstallCommand):

    short_description = 'Install backend'

    arguments = (
        Arg(
            'name', metavar='NAME', nargs='?', default='youtube_dl',
            help='Backend name. Default is youtube_dl.'
        ),
        Arg(
            'version', nargs='?', metavar='VERSION',
            help='Backend version. Default is latest.',
        ),
        Arg(
            '-f', '--force', action='store_true',
            help='Force installation if the same version is already installed.'
        ),
    )

    def get_output_dir(self, wheel: Wheel) -> Path:
        return backend.get_backend_dir(wheel.name)

    def get_name_version_tuple(self) -> Tuple[str, Optional[str]]:
        return (self.args.name, self.args.version)

    def get_force_flag(self) -> bool:
        return self.args.force



from dl_plus.cli.commands.base import CommandGroup

from .info import BackendInfoCommand
from .install import BackendInstallCommand


class BackendCommandGroup(CommandGroup):

    short_description = 'Backend management commands'

    commands = (
        BackendInfoCommand,
        BackendInstallCommand,
    )



import argparse
from typing import List

from .commands import RootCommandGroup
from .commands.base import CommandGroup


__all__ = ['run_command']


_COMMAND_DEST = 'command'


class CommandNamespace(argparse.Namespace):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__dict__[_COMMAND_DEST] = []

    def _get_command(self):
        return self.__dict__[_COMMAND_DEST]

    def _set_command(self, value):
        self.__dict__[_COMMAND_DEST].append(value)

    command = property(_get_command, _set_command)


class CommandArgParser(argparse.ArgumentParser):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault(
            'formatter_class', argparse.RawDescriptionHelpFormatter)
        super().__init__(*args, **kwargs)

    def add_command_arguments(self, command):
        for arg in command.arguments:
            arg.add_to_parser(self)

    def add_command_group(self, command_group):
        command_group_subparsers = self.add_command_group_subparsers(
            title=command_group.short_description)
        for command_or_group in command_group.commands:
            description = (
                command_or_group.long_description
                or command_or_group.short_description
            )
            command_parser = command_group_subparsers.add_parser(
                command_or_group.name,
                help=command_or_group.short_description,
                description=description,
            )
            if issubclass(command_or_group, CommandGroup):
                command_parser.add_command_group(command_or_group)
            else:
                command_parser.add_command_arguments(command_or_group)

    def add_command_group_subparsers(self, *args, **kwargs):
        kwargs.setdefault('dest', _COMMAND_DEST)
        kwargs.setdefault('metavar', 'COMMAND')
        kwargs.setdefault('parser_class', self.__class__)
        # Python 3.6 — add_subparsers(required=...) is not supported
        subparsers = self.add_subparsers(*args, **kwargs)
        subparsers.required = True
        return subparsers


def run_command(prog: str, cmd_arg: str, args: List[str]) -> None:
    parser = CommandArgParser(prog=f'{prog} {cmd_arg}')
    parser.add_argument(
        cmd_arg, action='store_true', required=True, help=argparse.SUPPRESS)
    parser.add_command_group(RootCommandGroup)
    parsed_args = parser.parse_args(args, namespace=CommandNamespace())
    command_cls = RootCommandGroup.get_command(parsed_args.command)
    command_cls(parsed_args).run()



from dl_plus.backend import init_backend
from dl_plus.cli import args
from dl_plus.cli.commands.base import Command


class BackendInfoCommand(Command):

    short_description = 'Show backend information'

    arguments = (
        args.dlp_config,
        args.backend,
    )

    def run(self):
        backend = self.args.backend or self.config.backend
        backend_info = init_backend(backend)
        print('import name:', backend_info.import_name)
        print('version:', backend_info.version)
        print('path:', str(backend_info.path))
        print('managed:', 'yes' if backend_info.is_managed else 'no')
        metadata = backend_info.metadata
        if metadata:
            print('project name:', metadata.name)
            print('project version:', metadata.version)



from dl_plus.cli.commands.base import CommandGroup

from .install import ExtractorInstallCommand


class ExtractorCommandGroup(CommandGroup):

    short_description = 'Extractors management commands'

    commands = (
        ExtractorInstallCommand,
    )



import sys
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Iterable, List, Set, Type

from dl_plus import ytdl
from dl_plus.config import get_config_home
from dl_plus.extractor import machinery
from dl_plus.extractor.peqn import PEQN


if TYPE_CHECKING:
    from .extractor.extractor import Extractor


extractor_plugins_dir = get_config_home() / 'extractors'


def get_extractor_plugin_dir(ns: str, plugin: str) -> Path:
    return extractor_plugins_dir / f'{ns}-{plugin}'


def get_extractors(names: Iterable[str]) -> List[Type['Extractor']]:
    extractors_dict: Dict[Type['Extractor'], bool] = {}
    added_search_paths: Set[str] = set()

    def maybe_add_search_path(path: Path) -> None:
        if not path.is_dir():
            return
        path_str = str(path)
        if path_str not in added_search_paths:
            sys.path.insert(0, path_str)
            added_search_paths.add(path_str)

    for name in names:
        if name == ':builtins:':
            extractors = ytdl.get_all_extractors(include_generic=False)
        elif name == ':plugins:':
            if extractor_plugins_dir.is_dir():
                for path in extractor_plugins_dir.iterdir():
                    maybe_add_search_path(path)
            extractors = machinery.load_all_extractors()
        elif '/' in name:
            peqn = PEQN.from_string(name)
            maybe_add_search_path(get_extractor_plugin_dir(
                peqn.ns, peqn.plugin))
            extractors = machinery.load_extractors_by_peqn(peqn)
        else:
            extractors = ytdl.get_extractors_by_name(name)
        for extractor in extractors:
            if extractor not in extractors_dict:
                extractors_dict[extractor] = True
    return list(extractors_dict.keys())


def enable_extractors(names) -> None:
    extractors = get_extractors(names)
    ytdl.patch_extractors(extractors)



import re
from pathlib import Path
from typing import Optional, Tuple

from dl_plus.cli.args import Arg
from dl_plus.cli.commands.base import BaseInstallCommand, CommandError
from dl_plus.core import get_extractor_plugin_dir
from dl_plus.pypi import Wheel


PLUGIN_NAME_REGEX = re.compile(
    r'^(?:dl-plus-extractor-)?(?P<ns>[a-z0-9]+)[/-](?P<plugin>[a-z0-9]+)$')


class ExtractorInstallCommand(BaseInstallCommand):

    short_description = 'Install extractor plugin'

    arguments = (
        Arg(
            'name', metavar='NAME',
            help='Extractor plugin name.'
        ),
        Arg(
            'version', nargs='?', metavar='VERSION',
            help='Extractor plugin version. Default is latest.',
        ),
        Arg(
            '-f', '--force', action='store_true',
            help='Force installation if the same version is already installed.'
        ),
    )

    ns: str
    plugin: str

    def init(self):
        plugin_name = self.args.name
        match = PLUGIN_NAME_REGEX.fullmatch(plugin_name)
        if not match:
            raise CommandError(f'Invalid extractor plugin name: {plugin_name}')
        self.ns, self.plugin = match.groups()

    def get_output_dir(self, wheel: Wheel) -> Path:
        return get_extractor_plugin_dir(self.ns, self.plugin)

    def get_name_version_tuple(self) -> Tuple[str, Optional[str]]:
        return (
            f'dl-plus-extractor-{self.ns}-{self.plugin}', self.args.version)

    def get_force_flag(self) -> bool:
        return self.args.force



import argparse
import os.path
import sys
from textwrap import dedent

from dl_plus import core, ytdl
from dl_plus.backend import init_backend
from dl_plus.config import Config
from dl_plus.const import DL_PLUS_VERSION
from dl_plus.exceptions import DLPlusException

from . import args


__all__ = ['main']

_PROG = 'dl-plus'
_CMD = '--cmd'


def _dedent(text):
    return dedent(text).strip()


def _is_running_as_youtube_dl(program_name: str) -> bool:
    return os.path.basename(program_name) in ['youtube-dl', 'youtube-dl.exe']


class _MainArgParser(argparse.ArgumentParser):

    def format_help(self):
        return super().format_help() + ytdl.get_help()


def _get_main_parser() -> argparse.ArgumentParser:
    parser = _MainArgParser(
        prog=_PROG,
        usage=(
            '%(prog)s '
            '[--dlp-config PATH | --no-dlp-config] '
            '[--backend BACKEND] [--extractor EXTRACTOR] '
            '[YOUTUBE-DL OPTIONS] URL [URL...]'
        ),
        description=_dedent("""
            %(prog)s is a youtube-dl extension with pluggable extractors.

            The following are %(prog)s options:
        """),
        epilog='The following are youtube-dl options:',
        add_help=False,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    args.dlp_config.add_to_parser(parser)
    args.backend.add_to_parser(parser)
    parser.add_argument(
        '--dlp-version',
        action='version',
        version=DL_PLUS_VERSION,
        help='Print dl-plus version and exit.',
    )
    extractor_group = parser.add_mutually_exclusive_group()
    extractor_group.add_argument(
        '-E', '--extractor',
        action='append',
        help=_dedent("""
            Extractor name. Can be specified multiple times: -E foo -E bar.
        """),
    )
    extractor_group.add_argument(
        '--force-generic-extractor',
        action='store_true',
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        '-h', '--help',
        action='store_true',
        help=argparse.SUPPRESS,
    )
    return parser


def _main(argv):
    args = argv[1:]
    if '-U' in args or '--update' in args:
        raise DLPlusException('update is not yet supported')
    compat_mode = _is_running_as_youtube_dl(argv[0])
    config = Config()
    backend = None
    if not compat_mode:
        if _CMD in args:
            from .command import run_command
            run_command(prog=_PROG, cmd_arg=_CMD, args=args)
            return
        parser = _get_main_parser()
        parsed_args, ytdl_args = parser.parse_known_args(args)
        backend = parsed_args.backend
        if not parsed_args.no_dlp_config:
            config.load(parsed_args.dlp_config)
    else:
        ytdl_args = args
        config.load()
    if not backend:
        backend = config['main']['backend']
    init_backend(backend)
    force_generic_extractor = False
    extractors = None
    if not compat_mode:
        if parsed_args.help:
            parser.print_help()
            return
        force_generic_extractor = parsed_args.force_generic_extractor
        extractors = parsed_args.extractor
    if force_generic_extractor:
        ytdl_args.append('--force-generic-extractor')
    else:
        if not extractors:
            extractors = config.options('extractors.enable')
        core.enable_extractors(extractors)
    backend_options = config.get_backend_options()
    if backend_options is not None:
        ytdl_args = ['--ignore-config'] + backend_options + ytdl_args
    ytdl.run(ytdl_args)


def main(argv=None):
    if argv is None:
        argv = sys.argv
    try:
        _main(argv)
    except DLPlusException as exc:
        sys.exit(exc)



class Arg:

    __slots__ = ('args', 'kwargs')

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def add_to_parser(self, parser):
        return parser.add_argument(*self.args, **self.kwargs)


class ExclusiveArgGroup:

    __slots__ = ('arguments', 'required')

    def __init__(self, *arguments, required=False):
        self.arguments = arguments
        self.required = required

    def add_to_parser(self, parser):
        group = parser.add_mutually_exclusive_group(required=self.required)
        for arg in self.arguments:
            arg.add_to_parser(group)


dlp_config = ExclusiveArgGroup(
    Arg('--dlp-config', metavar='PATH', help='dl-plus config path.'),
    Arg(
        '--no-dlp-config', action='store_true',
        help='Do not read dl-plus config.',
    ),
)

backend = Arg(
    '--backend',
    metavar='BACKEND',
    help='youtube-dl backend.',
)
